#!/bin/busybox sh
#
# When overlaying the sdcard over the internal /var/log
# folder, file name collision are resolved that only the file
# in the sdcard is shown, the file in the internal flash is hidden
# This script will check for such duplicates and move the file from
# the internal flash to the backup folder on the sdcard
#

PROG="$(basename "$0" .sh)"
BACKUP_DIR="$SDO_ROOT_DIR"/media/card/backup

# these variables can be overridden from the unit test
ROOTFS_MNTPOINT="$SDO_ROOT_DIR"/mnt/rootfs

usage() {
    echo "Usage:"
    echo "  $PROG [options]"
    echo "Options:"
    echo "  -x      execute, dry-run by default"
    echo "  -h      help"
    echo ""
    echo "Removes duplicate files from internal flash. Duplicate files are files"
    echo "that exist on both branches of the overlay mount. This can happen if the"
    echo "sd-card was not mounted for some time."
    echo ""
}

EXECUTE=

while getopts "hx" option; do
    case $option in
        x)
            EXECUTE=1
            ;;
        h)
            usage
            exit 0
            ;;
        \?)
            usage
            exit 1
            ;;
    esac

done

shift $(($OPTIND - 1))


collisions() {
    (cd "$1"; find . -type f -print) | \
    while read FILE; do
        if [ -f "$2"/"$FILE" ]; then
            echo "${FILE#./}"
        fi
    done
}

backup_file() {
    local FILE_PATH="$1" # ds485p/ds485p.log
    local FEATURE_PATH="$2" # var/log/

    local BKUP_DIR="$BACKUP_DIR"/"$("$SDO_ROOT_DIR"/bin/date "+%Y-%m-%d")"
    local DST_PATH="$BKUP_DIR"/"$FEATURE_PATH"/"$(dirname $FILE_PATH)"
    local SRC_PATH="$ROOTFS_MNTPOINT"/"$FEATURE_PATH"

    if [ ! -d "$DST_PATH" ]; then
        "$SDO_ROOT_DIR"/bin/mkdir -p "$DST_PATH"
    fi
    "$SDO_ROOT_DIR"/bin/mv "$SRC_PATH"/"$FILE_PATH" "$DST_PATH"
}

rootfs_dev()
{
    sed 's/.*root=\([^ ]*\).*/\1/' "$SDO_ROOT_DIR"/proc/cmdline
}

rootfs_mounted()
{
    #
    # mountpoint(1) does not work if root device is mounted
    # twice. it stat(2) '/' and '/mnt/rootfs' and compares
    # st_dev/st_ino, if they differs it's a mointpoint
    # in our case this will not work, since both times
    # it is the root device
    # http://code.metager.de/source/xref/gnu/sysvinit/src/mountpoint.c#119
    #
    grep -q "$(rootfs_dev)\s*${ROOTFS_MNTPOINT#$SDO_ROOT_DIR}\s" \
        "$SDO_ROOT_DIR"/etc/mtab
}

mount_rootfs()
{
    ROOTDEV="$(rootfs_dev)"
    if ! $(expr ${ROOTDEV} : '.*\(mtdblock\).*' > /dev/null); then
        echo "ERROR: invalid root device $ROOTDEV" >&2
        return 1
    fi

    ROOTFS_TYPE=$(sed 's/.*rootfstype=\([^ ]*\).*/\1/' "$SDO_ROOT_DIR"/proc/cmdline)
    if [ -n $ROOTFS_TYPE -a $ROOTFS_TYPE != 'jffs2' ]; then
        echo "ERROR: unsupported filesystem $ROOTFS_TYPE" >&2
        return 2
    fi

    if rootfs_mounted; then
        echo "already mounted"
        return 0
    fi

    if [ ! -d "$ROOTFS_MNTPOINT" ]; then
        "$SDO_ROOT_DIR"/bin/mkdir -p "$ROOTFS_MNTPOINT"
    fi

    "$SDO_ROOT_DIR"/bin/mount -t "$ROOTFS_TYPE" "$ROOTDEV" "$ROOTFS_MNTPOINT"
}

umount_rootfs()
{
    if rootfs_mounted; then
        "$SDO_ROOT_DIR"/bin/umount "$ROOTFS_MNTPOINT"
    fi
}

parse_overlays()
{
    #
    # technically unions consists of branches(br) of
    # directories, where the first directory is implicitly
    # writable and those that follow are read-only.
    # permission can be overridden, as is in our case
    # br=writable[=rw]:readable-dir[=ro]
    # NOTE: make sure to emit RO dir before RW directory
    #
    awk '/^aufs/ {print $4}' "$SDO_ROOT_DIR"/etc/fstab | \
        sed -n -e 's/.*br:\([[:graph:]]*\)=rw:\([[:graph:]]*\)=ro/\2 \1/p' \
        -e 's/.*br:\([[:graph:]]*\)=ro:\([[:graph:]]*\)=rw/\1 \2/p' \
        -e 's/.*br:\([[:graph:]]*\):\([[:graph:]]*\)/\2 \1/p'
}

# main
echo "$("$SDO_ROOT_DIR"/bin/date +"%Y-%m-%dT%H:%M:%S") $PROG starting"

if ! "$SDO_ROOT_DIR"/bin/mountpoint -q /media/card; then
    echo "sdcard not mounted"
    exit
fi

trap umount_rootfs EXIT HUP INT QUIT ABRT KILL TERM

if ! mount_rootfs; then
    exit 255
fi

parse_overlays | while read RO RW; do
    # access read-only branch via the double mounted rootfs
    collisions "$ROOTFS_MNTPOINT"/"$RO" "$SDO_ROOT_DIR"/"$RW" | sort | \
    while read COMMON_FILE; do
        echo "duplicate: ${RO}/${COMMON_FILE}"
        if [ -n "$EXECUTE" ]; then
            backup_file "$COMMON_FILE" "$RO"
        fi
    done
done

if [ -n "$EXECUTE" ]; then
    # remove coredumps from internal flash
    "$SDO_ROOT_DIR"/bin/rm -rvf "${ROOTFS_MNTPOINT}"/var/log/coredumps || true
fi


echo "$("$SDO_ROOT_DIR"/bin/date +"%Y-%m-%dT%H:%M:%S") $PROG done"
